import sympy
import syngular


class Ring(object):

    def __init__(self, field, variables, ordering):
        self.field = field
        self.variables = variables
        self.ordering = ordering

    def __hash__(self):
        return hash(str(self))

    @property
    def field(self):
        return self._field

    @field.setter
    def field(self, field):
        if isinstance(field, tuple) and isinstance(field[0], str):
            field = sympy.symbols(field)
        self._field = field

    @property
    def variables(self):
        return self._variables

    @variables.setter
    def variables(self, variables):
        if not isinstance(variables, tuple):
            raise Exception("Ring variables must be tuple")
        if len(variables) == 0:
            raise Exception("Ring must be defined over at least one variable")
        if isinstance(variables[0], str):
            variables = sympy.symbols(variables)
        self._variables = variables

    @property
    def ordering(self):
        """Monomial ordering"""
        return self._ordering

    @ordering.setter
    def ordering(self, ordering):
        if isinstance(ordering, str):
            assert ordering[:2] in ['lp', 'rp', 'dp', 'Dp', 'ls', 'rs', 'ds', 'Ds', 'wp']
            if ordering[1] == 's':
                print("Warning: chosen ordering is not a well-ordering.")
        else:
            ordering = str(ordering).replace("'", "")
        self._ordering = ordering

    def __str__(self):
        string = ", ".join(map(str, [self.field, self.variables, self.ordering])).replace(",)", ")")
        if syngular.DEGBOUND != 0:
            string += f";\ndegBound = {syngular.DEGBOUND};\noption()"
        return string

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        assert isinstance(self, Ring) and isinstance(other, Ring)
        return self.field == other.field and self.variables == other.variables and self.ordering == other.ordering

    def zero_ideal(self):
        """Returns the zero ideal ⟨0⟩ in the ring."""
        from .ideal import Ideal
        return Ideal(self, [])

    def random_point(self, field, seed=None):
        """Returns a random numerical point in the given field on the zero ideal of the ring."""
        j = self.zero_ideal()
        return j.point_on_variety(field=field, seed=seed)



import functools
import mpmath
import numpy
import random
import re
import sympy
import syngular
import warnings

from copy import copy, deepcopy
from packaging import version

from pyadic import ModP

from mpmath.libmp.libhyper import NoConvergence

from .tools import RootNotInFieldError, RootPrecisionError
from .field import Field
from .polynomial import Monomial, Polynomial
from .settings import TemporarySetting

if version.parse(sympy.__version__) < version.parse('1.14'):
    # See sympy issue #23861, fixed in sympy pull request #27650
    # this fixes a weird bug where sympy does not respect precision even if mpmath.mp.dps precision is set
    # (sympy seems to use mpmath as backhand)
    equation = sympy.sympify(f"x - 1.{'0' * 290}1")
    sympy.nroots(equation, n=300, maxsteps=500)


def retry_to_find_root(max_tries=100):
    def retry_to_find_root_decorator(func):
        @functools.wraps(func)
        def wrapper(self, field, base_point={}, directions=None, valuations=tuple(), indepSetNbr=None, indepSet='guess',
                    seed=None, verbose=False):

            if indepSetNbr is not None and indepSet == 'guess':
                indepSet = indepSetNbr
                warnings.warn(
                    "Key word argument 'indepSetNbr' is deprecated and will be removed in a future version. "
                    "Please use 'indepSet' instead.", DeprecationWarning, stacklevel=2
                )

            if indepSet == 'guess':  # check if indep set can be computed quickly
                if verbose:
                    print("At first try, trying to compute the indepSets. ", end="")
                with TemporarySetting("syngular", "TIMEOUT", 3):
                    try:
                        self.indepSets
                        indepSet = None  # if this suceeds there is no reason to guess
                        if verbose:
                            print("IndepSet computation is cheap. Will compute it.")
                    except TimeoutError:
                        if verbose:
                            print("Gave up after 3 seconds. Will proceed with guesses.")
            elif indepSet in ('force guess', 'force-guess'):
                indepSet = 'guess'

            if base_point != {} and indepSet not in [None, 'guess']:
                return func(self, field, base_point=base_point, directions=directions, valuations=valuations,
                            indepSet=indepSet, seed=seed, verbose=verbose)
            else:
                for try_nbr in range(max_tries):
                    try:
                        res = func(self, field, base_point=base_point, directions=directions, valuations=valuations,
                                   indepSet=indepSet, seed=seed, verbose=verbose)
                        break
                    except (RootNotInFieldError, RootPrecisionError, NoConvergence, AssertionError,
                            *((TimeoutError, ) if indepSet == "guess" else ())) as e:
                        if try_nbr != max_tries - 1:
                            if verbose:
                                print(f"Caught {type(e).__name__} at try number {try_nbr}, retrying...")
                            if seed is not None:  # maintain pseudo-randomness, but change seed, else retring has no effect.
                                random.seed(seed)
                                seed += random.randint(10 ** 5, 10**6)
                            continue
                        else:
                            raise type(e)(f"Could not find a solution in {field} after {max_tries} attempts. {e}")
                return res
        return wrapper
    return retry_to_find_root_decorator


class Variety_of_Ideal:

    @retry_to_find_root(max_tries=100)
    def point_on_variety(self, field, base_point={}, directions=None, valuations=tuple(), indepSet='guess',
                         seed=None, verbose=False, directions_analytic_check=False):
        """Generate a representative point on or close to the variety associated to this ideal.
        The point is 'valuations' away from the exact variety, in the directions specified by 'directions'.
        If 'directions' are not provided, pick the first n=codim simplest generators from 'self'.
        If the ideal is not prime, an irreducible branch will be picked at random."""

        from .ideal import Ideal
        from .qring import QuotientRing

        assert all([isinstance(valuation, str) or valuation >= 0 for valuation in valuations])
        original_self = self
        self = deepcopy(self)   # don't modify self within this fuanction

        random.seed(seed)

        # do not modify directions, in case re-try is triggered, better the input is identical
        directions = deepcopy(directions)
        # handle directions, i.e. the generators of the sub-ideal of maximal codimension
        if directions is not None and directions != [] and field.name == "finite field":
            if verbose:
                print("Warning: directions passed with a finite field. Discarding them.")
            directions = []
        if directions is None or directions == []:
            directions = []
            if field.name != "finite field":
                if verbose:
                    print("Directions not provided, obtaining them from ideal generators.")
                for poly in sorted(self.generators, key=lambda x: len(x)):
                    if Ideal(self.ring, directions + [poly, ]).codim > Ideal(self.ring, directions).codim:
                        directions += [poly, ]
                    if len(directions) == self.codim:
                        break
                assert Ideal(self.ring, directions).codim == self.codim
        elif directions_analytic_check:
            # make sure the provided directions make sense, given the ideal (they need to belong to it)
            # if this is not triggered, then the check is perfomed numerically later. The analytic check can be expensive.
            for direction in directions:
                if Polynomial(direction, Field("rational", 0, 0)) not in self:
                    raise Exception(f"Invalid direction, {direction} was not in {self}.")
        # make sure provided directions are expanded strings
        for i, direction in enumerate(directions):
            if isinstance(direction, sympy.core.Basic):
                directions[i] = str(sympy.expand(direction))
        # extra directions in qring
        if isinstance(self.ring, QuotientRing) and field.name not in ['finite field', 'Fp']:
            directions += self.ring.ideal.generators

        # handle valuations - if valuations == tuple() return a point exactly on the variety
        if field.name == "padic":
            prime, iterations = field.characteristic, field.digits
            if valuations == tuple():
                valuations = tuple(field.digits for _ in directions)
        elif field.name == "finite field":
            prime, iterations = field.characteristic, 1
        elif field.name == "mpc":
            prime, iterations = None, 1 if valuations == tuple() else 2
        # extra valuations in qring
        if isinstance(self.ring, QuotientRing) and field.name == "padic":
            valuations = valuations + (field.digits, ) * len(self.ring.ideal.generators)
        elif isinstance(self.ring, QuotientRing) and field.name == "mpc":
            valuations = valuations + (0, ) * len(self.ring.ideal.generators)

        if verbose:
            print("Directions:", directions, )
            print("Valuations:", valuations)

        # if in qring, go to full ring
        if isinstance(self.ring, QuotientRing):
            self.to_full_ring()

        # handle independent set
        if indepSet == 'guess':
            chose_indepSet = self.guess_indep_set()
        elif indepSet is None or isinstance(indepSet, int):
            indepSets = self.indepSets
            if verbose:
                print("Codimensions:", set(indepSet.count(0) for indepSet in indepSets))
                print("Number of indepSets:", len(indepSets))
            chose_indepSet = indepSets[random.randint(0, len(indepSets) - 1) if indepSet is None else indepSet]
        elif isinstance(indepSet, tuple):
            chose_indepSet = indepSet
        else:
            raise NotImplementedError
        indepSymbols = tuple([str(symbol) for i, symbol in enumerate(self.ring.variables) if chose_indepSet[i] == 1])
        depSymbols = tuple([str(symbol)for i, symbol in enumerate(self.ring.variables) if chose_indepSet[i] == 0])
        if verbose:
            print("Chosen indepSet:", chose_indepSet)
            print("indepSymbols:", indepSymbols)
            print("depSymbols:", depSymbols)
            print(f"{'Guessing codim' if indepSet == 'guess' and self._dim is None else 'Codim'} {len(depSymbols)} variety in {len(chose_indepSet)}-dim space")

        # handle the base point, i.e. the values of the independent variables
        if base_point == {}:
            base_point = {indepSymbol: field.random() for indepSymbol in indepSymbols}
        else:
            base_point = {str(key): field(val) for key, val in base_point.items()}
        base_point |= {depSymbol: Polynomial(depSymbol, field) for depSymbol in depSymbols}

        oSemiNumericalIdeal = self._semi_numerical_slice(field, directions, valuations, base_point, depSymbols, verbose=verbose, iteration=0)

        # print(repr(oSemiNumericalIdeal))

        for iteration in range(iterations):

            if verbose:
                print(f"\nAt iteration {iteration}")
                # print(f"base point {base_point}")
                # print(repr(oSemiNumericalIdeal), oSemiNumericalIdeal.primary_decomposition, oSemiNumericalIdeal.groebner_basis, len(oSemiNumericalIdeal.groebner_basis))

            syngular.DEGBOUND = 0

            # and oSemiNumericalIdeal.dim == -1:  # this is the ideal generated by '1'
            # check it explicitly, unless the .dim property is reverted to use grobner_basis instead of std
            if prime is None and oSemiNumericalIdeal.groebner_basis == ['1']:
                raise RootPrecisionError
            if not oSemiNumericalIdeal.dim == 0:
                if oSemiNumericalIdeal.dim > 0 and indepSet == 'guess':
                    # determines dimension of original ideal in the full ring from that of the semi-numerical slice
                    learnt_dimension = len(self.ring.variables) - len(self.generators) + oSemiNumericalIdeal.dim
                    if isinstance(original_self.ring, QuotientRing):
                        original_self.dim_in_full_ring = learnt_dimension
                    else:
                        original_self.dim = learnt_dimension
                    if verbose:
                        print(f"Determined the actual dimension to be {self.dim}")
                raise AssertionError(f"The dimension of the semi-numerical ideal was {oSemiNumericalIdeal.dim} instead of zero.")

            root_dicts = lex_groebner_solve(oSemiNumericalIdeal.groebner_basis, prime=prime)
            check_solutions(oSemiNumericalIdeal.groebner_basis, root_dicts, field)  # they may be stricter then wanted for mpc.

            try:
                root_dict = root_dicts[0]
            except IndexError:
                if not field.is_algebraically_closed:
                    raise RootNotInFieldError(f"Got root_dicts: {root_dicts}, for lex Groebner basis:\n{oSemiNumericalIdeal.groebner_basis}.")
                else:
                    raise IndexError(f"Got root_dicts: {root_dicts}, for lex Groebner basis:\n{oSemiNumericalIdeal.groebner_basis}.")
            # root_dict = {key: root_dict[key] for key in root_dict.keys() if key not in indepSymbols}

            if iteration < iterations - 1:
                for key in root_dict.keys():
                    if field.name == "padic":
                        root_as_poly = Polynomial(root_dict[key], Field("finite field", field.characteristic, 1))
                        root_as_poly.field = field
                    else:
                        root_as_poly = Polynomial(root_dict[key], field)
                    root_dict[key] = root_as_poly + Polynomial(f"{(prime if prime is not None else 1)} * {key}", field)

            update_point_dict(base_point, root_dict, field)
            # print("updated point:", base_point)

            if iteration == 0 and not directions_analytic_check:
                # instead of analytically checking the directions are consistent with the ideal
                # we check they vanish numerically at the constructed point
                for direction in directions:
                    num_poly = Polynomial(direction, field).subs(base_point).subs({key: 0 for key in depSymbols}).coeffs[0]
                    # print("val:", float(abs(num_poly)))
                    if abs(num_poly) > abs(field.ε):
                        raise Exception(f"Invalid direction, {direction} was not in {self}. Numerical membership check failed.")

            if iteration < iterations - 1:
                if prime is not None:
                    valuations = [valuation - 1 for valuation in valuations]
                    valuations_at_directions = list(filter(lambda x: x[1] > 0, zip(directions, valuations)))
                else:
                    valuations_at_directions = list(zip(directions, valuations))
                if valuations_at_directions != []:
                    directions, valuations = zip(*valuations_at_directions)
                else:
                    directions, valuations = (), ()
                if verbose:
                    print("New directions, valuations:", directions, valuations)

                oSemiNumericalIdeal = self._semi_numerical_slice(field, directions, valuations, base_point, depSymbols, verbose=False, iteration=iteration + 1)

                if len(oSemiNumericalIdeal.indepSets) == 1 and numpy.all(numpy.array(oSemiNumericalIdeal.indepSets) == 0):
                    continue

                currentIndepSet = oSemiNumericalIdeal.indepSets[0]
                newIndepSymbols = tuple([symbol for i, symbol in enumerate(depSymbols) if currentIndepSet[i] == 1])

                if verbose:
                    print("New independent symbols:", newIndepSymbols)

                if newIndepSymbols != tuple():  # this happens only with padics - codimension during iterative lift may drop

                    if field.name != "padic":
                        raise Exception(f"Codimension changed while not using padics, the field was {field}. Are you sure number of valuations is correct?")

                    rand_dict = {newIndepSymbol: random.randrange(1, field.characteristic ** (field.digits - iteration)) for newIndepSymbol in newIndepSymbols}

                    update_point_dict(base_point, rand_dict, field)

                    depSymbols = tuple(symbol for symbol in depSymbols if symbol not in newIndepSymbols)

                    if depSymbols == tuple():  # no more equations to solve, terminate early
                        break

                    oSemiNumericalIdeal = self._semi_numerical_slice(field, directions, valuations, base_point, depSymbols, verbose=False, iteration=iteration + 1)

        for key, val in base_point.items():
            if val not in field:
                if not isinstance(base_point[key], Polynomial):
                    raise ValueError(f"{base_point[key]} is {type(base_point[key])} instead of a Polynomial")
                assert len(base_point[key]) == 1 and base_point[key].coeffs_and_monomials[0][1] == Monomial("")
                base_point[key] = base_point[key].coeffs_and_monomials[0][0]

        return {str(key): val for key, val in base_point.items()}

    def _semi_numerical_slice(self, field, directions, valuations, base_point, depSymbols, verbose=False, iteration=0):
        """Helper function for point_on_variety. Uses the values in 'base_point' to return a new ideal of lower dimension.
        The potentially perturbed slice of the origial ideal 'self'."""

        from .ideal import Ideal
        from .ring import Ring

        # on subsequent iterations, switch to an ideal of maximal codimension (potentiallty losing branch information), beacuse:
        # (floats) need to append perturbations to equations; (padics) may need to solve less equations; (finite field) iteration > 0 does not make sense, perturbation impossible.
        if iteration > 0:
            generators = list(copy(directions))
            if field.characteristic == 0:
                for i, valuation in enumerate(valuations):
                    generators[i] = str(generators[i]) + " + " + str(-field(valuation))
        else:
            generators = copy(self.generators)

        generators = [Polynomial(generator, field=field) for generator in generators]
        generators = [generator.subs(base_point, field=field) for generator in generators]
        generators = list(filter(lambda x: x != 0, generators))

        if field.characteristic == 0:
            generators = [re.sub(r"(\d)j", r"\1*I", str(generator), ) for generator in generators]
        elif field.name == "padic":
            for i, generator in enumerate(generators):
                generators[i] = generator / field.characteristic ** iteration
                generators[i].coeffs = [coeff.as_tuple_from_zero[0] for coeff in generators[i].coeffs]
                generators[i]._field = Field("finite field", field.characteristic, 1)
            generators = list(filter(lambda x: x != 0, generators))

        oZeroDimIdeal = Ideal(Ring(field.singular_notation, depSymbols, "lp"), generators)

        if verbose:
            print("Constructed semi-numerical ideal slice:")
            print(repr(oZeroDimIdeal))

        return oZeroDimIdeal


def update_point_dict(base_point_dict, new_vals_dict, field):
    for key in base_point_dict:
        if key in new_vals_dict and base_point_dict[key] == key:
            base_point_dict[key] = new_vals_dict[key]
        elif key in new_vals_dict:
            base_point_dict[key] = base_point_dict[key].subs(new_vals_dict)
        elif key in list(map(str, new_vals_dict.keys())):
            base_point_dict[key] = base_point_dict[key].subs({str(key): val for key, val in new_vals_dict.items()})
    return base_point_dict


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
# Univariate solvers
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def univariate_floating_point_solver(equation, root_dict):
    """Returns all possible solutions of 'equation' over arbitrary precision complex numbers."""
    equation = sympy.sympify(equation).subs(root_dict)
    free_symbols = list(equation.free_symbols)
    assert len(free_symbols) == 1
    symbol = free_symbols[0]
    solutions = list(map(mpmath.mpc, sympy.nroots(equation, n=300, maxsteps=500)))  # mpmath.polyroots is faster, but the parsing is more complicated
    return update_root_dict(symbol, solutions, root_dict)


def univariate_finite_field_solver(equation, root_dict, prime):
    """Returns all possible solutions of 'equation' over a finite field of cardinality 'prime'.
       If already satisfied returns True, if no solution exists returns False."""
    equation = sympy.sympify(equation).subs(root_dict)
    if isinstance(equation, sympy.core.numbers.Integer) and equation % prime == 0:
        return True
    free_symbols = list(equation.free_symbols)
    if len(free_symbols) < 1:
        return False
    if len(free_symbols) > 1:
        raise Exception("Too many free parameters.")
    symbol = free_symbols[0]
    equation = sympy.poly(equation, modulus=prime)
    if equation == 0:
        return True
    pre_factor, factors = sympy.factor_list(sympy.factor(equation, modulus=prime))
    factors = [factor[0] for factor in factors]
    if pre_factor % prime == 0:
        return True
    linear_factors = [factor for factor in factors if sympy.diff(factor, symbol) == 1]
    if linear_factors == []:
        return False
    solutions = [ModP(int(sympy.solve(factor)[0]), prime) for factor in linear_factors]
    return update_root_dict(symbol, solutions, root_dict)


def update_root_dict(symbol, solutions, root_dict):
    """Given solutions and root_dict returns updated root_dicts."""
    root_dicts = [deepcopy(root_dict)]
    root_dicts[0].update({symbol: solutions[0]})
    for solution in solutions[1:]:
        new_root_dict = deepcopy(root_dict)
        new_root_dict.update({symbol: solution})
        root_dicts.append(new_root_dict)
    return root_dicts


def lex_groebner_solve(equations, prime=None):
    """Returns the variety corresponding to a given zero dimensional ideal in lexicographic groebner basis form.
       The variety take the form of a list of dictionaries for the possible values of the variables."""
    root_dicts = [{}]
    for equation in equations:
        temp_dicts = []
        for i, root_dict in enumerate(root_dicts):
            if prime is None:
                sols = univariate_floating_point_solver(equation, root_dict)
            else:
                sols = univariate_finite_field_solver(equation, root_dict, prime)
            if sols is True:
                temp_dicts += [root_dict]
            elif sols is False:
                continue
            else:
                temp_dicts += sols
        else:
            root_dicts = temp_dicts
    return root_dicts


def check_solutions(equations, root_dicts, field):
    """Checks that all solutions in root_dicts solve the equations."""
    field = field if field.name not in ["padic", "Qp"] else Field("finite field", field.characteristic, 1)
    for root_dict in root_dicts:
        check = [Polynomial(equation, field).subs(root_dict) for equation in equations]
        assert all([len(entry) == 1 for entry in check])  # check these are elements of the field
        if not all([abs(entry.coeffs[0]) <= field.tollerance for entry in check]):
            if field.characteristic == 0:
                raise RootPrecisionError
            else:
                raise AssertionError



import re
import subprocess
import syngular
import random
import warnings

from pathlib import Path
from packaging.version import Version, InvalidVersion


def execute_singular_command(singular_command, timeout='default', verbose=False):
    if timeout == 'default':
        timeout = syngular.TIMEOUT
    if isinstance(singular_command, list):
        singular_command = "\n".join(singular_command)
    # print(singular_command)
    if len(singular_command) >= 131072:  # 2 ** 17
        Path("/tmp/.singular_commands/").mkdir(parents=True, exist_ok=True)
        random_integer = random.randint(0, 2**64 - 1)
        file_path = f"/tmp/.singular_commands/singular_command_{random_integer}"
        with open(file_path, "w+") as file:
            file.write(singular_command)
        test = subprocess.Popen(["timeout", "--verbose", str(timeout), "Singular", "--quiet", file_path],
                                stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    else:
        test = subprocess.Popen(["timeout", "--verbose", str(timeout), "Singular", "--quiet", "--execute", singular_command],
                                stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    try:
        output, stderr = test.communicate()
    except KeyboardInterrupt:
        print("Keyboard interrupt received. Terminating the Singular process.")
        if test.poll() is None:  # Check if the process is still running
            import os
            import signal
            os.kill(test.pid, signal.SIGTERM)
        raise KeyboardInterrupt
    output = output.decode("utf-8")
    if stderr is not None:
        stderr = stderr.decode("utf-8")
    if len(output) == 0 and stderr is None:
        raise SingularException(f"Empty output while executing: {singular_command}")
    if len(output) > 0 and output[-1] == "\n":
        output = output[:-1]
    if 'halt' in output or (stderr is not None and 'timeout' in stderr):
        raise TimeoutError(f"{timeout} s")
    if 'error' in output and 'groebner base computations with inexact coefficients can not be trusted due to rounding errors' not in output:
        raise SingularException(f"{output}\n\n\nError occured while executing:\n{singular_command}")
    if '   ? `Q` is undefined' in output:
        output = output.replace("   ? `Q` is undefined", "")
        print("Singular Error: Q is undefined")
    if '//options: redSB degBound redefine usage prompt\n' in output:
        output = output.replace("//options: redSB degBound redefine usage prompt\n", "")
        print("Singular Warning: redSB degBound redefine usage prompt")
    output = output.replace('// ** minbase applies only to the local or homogeneous case over coefficient fields\n', '')
    output = output.replace('//options: redSB degBound redTail redThrough intStrategy redefine usage prompt\n', '')
    output = output.replace('//options: degBound redTail redThrough intStrategy redefine usage prompt\n', '')
    output = re.sub(r'// \*\* .* is no standard basis\n', '', output)
    if 'groebner base computations with inexact coefficients can not be trusted due to rounding errors' in output:
        if verbose:
            print('Singular Warning: groebner base computations with inexact coefficients can not be trusted due to rounding errors')
        output = output.replace('// ** groebner base computations with inexact coefficients can not be trusted due to rounding errors\n', '')
    if len(singular_command) >= 131072:
        Path(file_path).unlink()
    if syngular.DEBUG:
        print("DEBUG - Command received:\n", singular_command)
        print("DEBUG - Output obtained:\n", output)
    return output


class SingularException(Exception):
    pass


class RootNotInFieldError(Exception):
    pass


class RootPrecisionError(Exception):
    pass


try:
    test = subprocess.Popen(["timeout", "5", "Singular", "--dump-versiontuple"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    output = test.communicate()[0]
    output = output.decode("utf-8").replace("\n", "")
    try:
        Singular_version = Version(output)
    except InvalidVersion:
        warnings.warn("Could not determine Singular version. Are you sure Singular is installed?", stacklevel=2)
        Singular_version = None
except FileNotFoundError as e:
    warnings.warn(f"\nAre you sure timeout is installed for use in the command line?\n{e}\nCould not determine Singular version.", stacklevel=2)
    Singular_version = None



import numpy
import sympy

from syngular import Ideal, Ring, QRing, Field, Polynomial


def test_instantiation():
    ring0 = Ring('0', sympy.symbols(('x1', 'x2')), 'dp')
    ring1 = Ring('0', ('x1', 'x2'), 'dp')
    ring2 = Ring('0', ('x', 'y'), 'dp')
    assert ring0 == ring1
    assert not ring0 == ring2


def test_qring_and_point_in_ring():
    m, D = 6, 7
    η = numpy.diag([1, ] + [-1, ] * (D - 1))
    momenta = numpy.array([numpy.array([sympy.symbols(f"p{i}_{j}") for j in range(D)]) for i in range(1, m + 1)])
    on_shell_relations = [momentum @ η @ momentum for momentum in momenta]
    momentum_conservation = sum(momenta).tolist()
    r = Ring('0', tuple(momenta.flatten().tolist()), 'dp')
    i = Ideal(r, on_shell_relations + momentum_conservation)
    q = QRing(r, i)
    field = Field('finite field', 2 ** 31 - 1, 1)
    point = q.random_point(field=field)
    assert all([Polynomial(generator, field)(point) == 0 for generator in i.generators])



import mpmath
import random

from fractions import Fraction as Q

from pyadic import PAdic, ModP
from pyadic.padic import padic_sqrt
from pyadic.finite_field import finite_field_sqrt

mpmath.mp.dps = 300


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


class Field(object):
    """A class representing number fields."""

    def __init__(self, *args):
        """Defaults to ('mpc', 0, 300)"""
        if args == ():
            args = ('mpc', 0, 300)
        self.set(*args)

    def set(self, *args):
        """(name, characteristic, digits)"""
        self.name = args[0]
        self.characteristic = args[1]
        self.digits = args[2]

    def __str__(self):
        return "('{}', {}, {})".format(self.name, self.characteristic, self.digits)

    def __repr__(self):
        return str(self)

    def __call__(self, other):
        """Cast to field."""
        if self.name in ["mpc", "C"]:
            return mpmath.mpc(mpmath.mpmathify(other))
        elif self.name in ["padic", "Qp"]:
            return PAdic(other, self.characteristic, self.digits)
        elif self.name in ["finite field", "Fp"]:
            return ModP(other, self.characteristic)
        elif self.name in ["rational", "Q"]:
            return Q(other)
        else:
            raise NotImplementedError

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value not in ['rational', 'Q', 'mpc', 'C', 'gaussian rational', 'finite field', 'Fp', 'padic', 'Qp']:
            raise Exception("Field must be one of 'mpc', 'gaussian rational', 'finite field', 'padic'.")
        else:
            self._name = value

    @property
    def characteristic(self):
        return self._characteristic

    @characteristic.setter
    def characteristic(self, value):
        if value < 0:
            raise Exception("Characteristic must be non-negative.")
        else:
            self._characteristic = value

    @property
    def digits(self):
        if self.name in ['mpc', 'C']:
            return mpmath.mp.dps
        else:
            return self._digits

    @digits.setter
    def digits(self, value):
        if value < 0:
            raise Exception("Digits must be positive.")
        elif self.name in ['mpc', 'C']:
            mpmath.mp.dps = value
        else:
            self._digits = value

    def __contains__(self, other):
        return isinstance(other, self(1).__class__)

    @property
    def is_algebraically_closed(self):
        if self.name in ['mpc', 'C']:
            return True
        else:
            return False

    @property
    def tollerance(self):
        if self.name in ['gaussian rational', 'finite field']:
            return 0
        elif self.name in ['mpc', 'C']:
            return mpmath.mpf('10e-{}'.format(int(min([0.95 * mpmath.mp.dps, mpmath.mp.dps - 4]))))
        elif self.name in ['padic', 'Qp']:
            return PAdic(0, self.characteristic, 0, self.digits)

    @property
    def singular_notation(self):
        if self.name in ['mpc', 'C']:
            return '(complex,{},I)'.format(self.digits - 5)
        elif self.name in ['finite field', 'Fp', 'padic', 'Qp']:
            return str(self.characteristic)
        else:
            return None

    def sqrt(self, val):
        if self.name in ["finite field", 'Fp']:
            if not isinstance(val, ModP):
                val = self(val)
            return finite_field_sqrt(val)
        elif self.name in ["padic", 'Qp']:
            if not isinstance(val, PAdic):
                val = self(val)
            return padic_sqrt(val)
        elif self.name in ["mpc", 'C']:
            return mpmath.sqrt(val)
        else:
            raise Exception(f"Field not understood: {self.field.name}")

    @property
    def j(self):
        return self.sqrt(-1)

    i = I = j

    def random(self, shape=(1, )):
        if shape == (1, ):
            if self.name in ["padic", 'Qp']:
                p, k = self.characteristic, self.digits
                return PAdic(random.randrange(0, p ** k - 1), p, k)
            elif self.name in ["finite field", 'Fp']:
                p = self.characteristic
                return ModP(random.randrange(0, p), p)
            elif self.name in ["mpc", 'C']:
                return mpmath.mpc(str(Q(random.randrange(-100, 101), random.randrange(1, 201))),
                                  str(Q(random.randrange(-100, 101), random.randrange(1, 201))))
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

    def random_square(self):
        from pyadic.field_extension import FieldExtension
        for i in range(10000):
            val = self.random()
            if not isinstance(self.sqrt(val), FieldExtension):
                return val

    def random_element(self, *args, **kwargs):
        import warnings
        warnings.warn(
            "random_element is deprecated and will be removed in a future version. "
            "Use random instead.",
            DeprecationWarning,
            stacklevel=2
        )
        return self.random(*args, **kwargs)

    def epsilon(self, shape=(1, ), ):
        if shape == (1, ):
            if not hasattr(self, "_ε"):
                if self.name == "padic":
                    self._ε = self.characteristic
                elif self.name == "finite field":
                    raise ValueError("Finite field infinitesimal does not exist.")
                elif self.name == "mpc":
                    self._ε = mpmath.mpf('1e-30')
                else:
                    raise NotImplementedError
            return self._ε
        else:
            raise NotImplementedError

    @property
    def ε(self):
        if not hasattr(self, "_ε"):
            self._ε = self.epsilon()
        return self._ε

    @ε.setter
    def ε(self, temp_ε):
        self._ε = temp_ε


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


field = Field()



import functools
import re
import sympy
import numpy
import inspect
import random

from .tools import execute_singular_command
from .ring import Ring
from .qring import QuotientRing
from .ideal_algorithms import Ideal_Algorithms
from .variety import Variety_of_Ideal
from .polynomial import Polynomial
from .field import Field


class Ideal(Ideal_Algorithms, Variety_of_Ideal, object):

    def __init__(self, ring, generators):
        if isinstance(generators, tuple):
            generators = list(generators)
        if not isinstance(ring, Ring) or not isinstance(generators, list):
            raise Exception("Invalid Ideal initialisation.")
        self.ring = ring
        self.generators = generators
        self.test_valid_ideal()
        self._dim = None

    def test_valid_ideal(self):
        singular_commands = [f"ring r = {self.ring};",
                             f"ideal i = {self};",
                             "print(i);"
                             "$"]
        execute_singular_command(singular_commands)

    @property
    def generators(self):
        return self._generators

    @generators.setter
    def generators(self, generators):
        if not isinstance(generators, list):
            raise Exception("Ideal generators must be list")
        if len(generators) == 0:
            generators = ['0']
        self._generators = generators

    def __hash__(self):
        return hash(", ".join(self.groebner_basis)) + hash(self.ring)

    def __eq__(self, other):
        return self.groebner_basis == other.groebner_basis

    def squoosh(self):
        self.generators = self.minbase

    @property
    def dim(self):
        if self._dim is None:
            singular_commands = [f"ring r = {self.ring};",
                                 f"ideal gb = {','.join(self.groebner_basis)};",
                                 # f"ideal i = {self};",   # Check which is better
                                 # "ideal gb = std(i);",   # {','.join(self.groebner_basis)};",
                                 "print(dim(gb));",
                                 "$"]
            output = execute_singular_command(singular_commands)
            self._dim = int(output)
        return self._dim

    @dim.setter
    def dim(self, val):
        self._dim = val

    @property
    def dims(self):
        return {entry.count(1) for entry in self.indepSets}

    @property
    def codim(self):
        return Ideal(self.ring, ['0']).dim - self.dim

    @property
    def codims(self):
        zeroIdeal_dim = Ideal(self.ring, ['0']).dim
        return {zeroIdeal_dim - dim for dim in self.dims}

    def guess_indep_set(self):
        equations = self.generators
        if isinstance(self.ring, QuotientRing):
            equations += self.ring.ideal.generators
        n = len(self.ring.variables)
        m = self.dim if self._dim is not None else (n - len(equations))
        lst = [0] * (n - m) + [1] * m
        equations_variables = [Polynomial(equation, field=Field("rational", 0, 0)).variables for equation in equations]
        for _ in range(1000):  # discard obviously wrong indep sets: all equations must have at least 1 dependent variable
            random.shuffle(lst)
            guess_dict = dict(zip(map(str, self.ring.variables), lst))
            if all([0 in [guess_dict[variable] for variable in variables] for variables in equations_variables]):
                break
        return tuple(lst)

    @functools.cached_property
    def indepSet(self):
        singular_commands = [f"ring r = {self.ring};",
                             f"ideal gb = {','.join(self.groebner_basis)};",
                             "print(indepSet(gb));",
                             "$"]
        output = execute_singular_command(singular_commands)
        return tuple(map(int, output.split(",\n")))

    @functools.cached_property
    def indepSets(self):
        singular_commands = [f"ring r = {self.ring};",
                             # f"ideal gb = {','.join(self.groebner_basis)};",   # this breaks singular variety construction, especially with mpcs
                             f"ideal i = {self};",
                             "ideal gb = groebner(i);",
                             "print(indepSet(gb, 1));",
                             "$"]
        output = execute_singular_command(singular_commands)
        if output == 'empty list':
            return [self.indepSet]
        indepSets = [tuple(map(int, line.replace(" ", "").split(","))) for line in output.split("\n") if ":" not in line]
        return indepSets

    @functools.cached_property
    def groebner_basis(self):
        singular_commands = ["option(redSB);",
                             f"ring r = {self.ring};",
                             f"ideal i = {self};",
                             "ideal gb = groebner(i);",
                             "short=0;",
                             "print(gb);",
                             "$"]
        singular_command = "\n".join(singular_commands)
        output = execute_singular_command(singular_command)
        output = [line.replace(",", "") for line in output.split("\n")]
        return output

    @functools.cached_property
    def leadGBmonomials(self):
        """Gives the leading monomials of the Groebner basis polynomials."""
        singular_commands = [f"ring r = {self.ring};",
                             f"ideal gb = {','.join(self.groebner_basis)};",
                             "def lts = lead(gb);",
                             "short=0;",
                             "print(lts);"]
        output = execute_singular_command(singular_commands)
        # print(output)
        output = [line.replace(",", "") for line in output.split("\n")]
        return output

    @functools.cached_property
    def minbase(self):
        singular_commands = [f"ring r = {self.ring};",
                             f"ideal i = {self};",
                             "short=0;",
                             "print(minbase(i));",
                             "$"]
        output = execute_singular_command(singular_commands)
        output = [line.replace(",", "") for line in output.split("\n")]
        return output

    @functools.cached_property
    def radical(self):
        """Returns the radical of the ideal."""
        singular_commands = ["LIB \"primdec.lib\";",
                             f"ring r = {self.ring};",
                             f"ideal i = {self};",
                             "def pr = radical(i);",   # options: GTZ / SY
                             "short=0;",
                             "print(pr);",
                             "$"]
        output = execute_singular_command(singular_commands)
        output = [line.replace(",", "") for line in output.split("\n")]
        cls, ring = self.__class__, self.ring
        return cls(ring, output)

    @functools.cached_property
    def primary_decomposition(self):
        singular_commands = ["LIB \"primdec.lib\";",
                             f"ring r = {self.ring};",
                             f"ideal i = {self};",
                             # f"ideal gb = {','.join(self.groebner_basis)};",
                             "def pr = primdecGTZ(i);",   # options: GTZ / SY
                             "short=0;",
                             "print(pr);",
                             "$"]
        output = execute_singular_command(singular_commands)
        output = [line.replace(",", "") for line in output.split("\n")]

        def clean_up(string):
            string = re.sub(r"_\[\d+\]=", "", string)
            string = re.sub(r"\[\d+\]:", "|", string)
            return string.replace(" ", "")

        primary_decomposed = ",".join(list(map(clean_up, output))).replace("|,", "|").replace(",|", "|").replace("|,", "|").split("||")
        primary_decomposed = list(filter(lambda x: x != '', primary_decomposed))
        primary_decomposed = [entry.split("|") for entry in primary_decomposed]
        primary_decomposed = [(entry[0].split(","), entry[1].split(",")) for entry in primary_decomposed]

        cls, ring = self.__class__, self.ring
        for i, (primary, prime) in enumerate(primary_decomposed):
            primary_decomposed[i] = (cls(ring, primary), cls(ring, prime))

        return primary_decomposed

    def eliminate(self, var_range):
        singular_commands = [f"ring r1 = {self.ring};",
                             f"ideal i = {self};",
                             f"ideal j = eliminate(i, {var_range.start} .. {var_range.stop});",
                             "print(minbase(j));",
                             "$"]
        output = execute_singular_command(singular_commands)
        output = [line.replace(",", "") for line in output.split("\n")]
        cls, ring = self.__class__, self.ring
        ring = Ring(ring.field, ring.variables[:var_range.start] + ring.variables[var_range.stop:], ring.ordering)
        return cls(ring, output)

    def __contains__(self, other):
        """Implements ideal membership."""
        if isinstance(other, Ideal):
            return self.__ideal_contains__(other)
        elif isinstance(other, (str, sympy.Expr, Polynomial)):
            return self.__poly_contains__(other)
        else:
            raise NotImplementedError

    def __ideal_contains__(self, other):
        qIdeal = self / other
        if list(map(str, qIdeal.generators)) == ['1']:
            return True
        else:
            return False

    def __poly_contains__(self, other):
        if not isinstance(other, str):
            other = str(other)
        assert isinstance(other, str)
        singular_commands = [f"ring r = {self.ring};",
                             f"ideal gb = {','.join(self.groebner_basis)};",
                             "poly f = " + other + ";",
                             "print(reduce(f, gb));"
                             "$"]
        output = execute_singular_command(singular_commands)
        output = [line.replace(",", "") for line in output.split("\n")]
        if output == ['0']:
            return True
        else:
            return False

    def __add__(self, other):
        """Addition of Ideals = Intersection of Varieties."""
        singular_commands = [f"ring r = {self.ring};",
                             f"ideal i = {self};",
                             f"ideal j = {other};",
                             "ideal k = i + j;",
                             "print(k);",
                             "$"]
        output = execute_singular_command(singular_commands)
        output = [line.replace(",", "") for line in output.split("\n")]
        cls, ring = self.__class__, self.ring
        return cls(ring, output)

    def __mul__(self, other):
        """Product of Ideals"""
        singular_commands = [f"ring r = {self.ring};",
                             f"ideal i = {self};",
                             f"ideal j = {other};",
                             "ideal k = i * j;",
                             "print(k);",
                             "$"]
        output = execute_singular_command(singular_commands)
        output = [line.replace(",", "") for line in output.split("\n")]
        cls, ring = self.__class__, self.ring
        return cls(ring, output)

    def __pow__(self, n):
        assert type(n) is int and n >= 0
        cls, ring = self.__class__, self.ring
        if n == 0:
            return cls(ring, ['1'])
        elif n % 2 == 0:
            root_2_res = self ** (n // 2)
            return root_2_res * root_2_res
        else:
            return self * (self ** (n - 1))

    def __and__(self, other):
        """Intersection of Ideals = Union of Varieties - This uses Python's set intersection operator '&'."""
        singular_commands = [f"ring r = {self.ring};",
                             f"ideal i = {self};",
                             f"ideal j = {other};",
                             "ideal k = intersect(i, j);",
                             "print(k);",
                             "$"]
        output = execute_singular_command(singular_commands)
        output = [line.replace(",", "") for line in output.split("\n")]
        cls, ring = self.__class__, self.ring
        return cls(ring, output)

    @staticmethod
    def intersection(*args):
        """Intersection of Ideals - wrapper around & operator for chained intersection."""
        if len(args) == 1:
            return args[0]
        res = args[0] & args[1]
        for arg in args[2:]:
            res = res & arg
        return res

    def __truediv__(self, other):
        """Quotient of ideals."""
        singular_commands = [f"ring r = {self.ring};",
                             f"ideal i = {self};",
                             f"ideal j = {other};",
                             "ideal k = quotient(i, j);",
                             "print(k);",
                             "$"]
        output = execute_singular_command(singular_commands)
        output = [line.replace(",", "") for line in output.split("\n")]
        cls, ring = self.__class__, self.ring
        return cls(ring, output)

    def reduce(self, other):
        """Remainder of division, i.e. reduction."""
        if isinstance(other, Ideal):
            singular_commands = [f"ring r = {self.ring};",
                                 f"ideal gb = {','.join(self.groebner_basis)};",
                                 f"ideal i = {other};",
                                 "ideal j = reduce(i, gb);",
                                 "print(j);",
                                 "$"]
            output = execute_singular_command(singular_commands)
            output = output.replace("\n", "").split(",")
            cls, ring = self.__class__, self.ring
            return cls(ring, output)
        else:
            singular_commands = [f"ring r = {self.ring};",
                                 f"ideal gb = {','.join(self.groebner_basis)};",
                                 f"poly f = {other};",
                                 "poly g = reduce(f, gb);",
                                 "print(g);",
                                 "$"]
            output = execute_singular_command(singular_commands)
            return output

    def __str__(self):
        return ",".join(map(str, self.generators))

    def __repr__(self):
        return "Ideal over Ring\n    " + repr(self.ring) + "\ngenerated by\n    " + str(self)

    def to_qring(self, other):
        if self._dim is not None:
            dim_in_full_ring = self._dim
            self.dim_in_full_ring = dim_in_full_ring
        self.delete_cached_properties()
        qring = QuotientRing(self.ring, other)
        self.ring = qring

    def to_full_ring(self):
        if isinstance(self.ring, QuotientRing):
            self.delete_cached_properties()
            self.generators = (self.ring.ideal + self).generators
            self.ring = self.ring.ideal.ring
            if hasattr(self, 'dim_in_full_ring'):
                self.dim = self.dim_in_full_ring
                delattr(self, 'dim_in_full_ring')

    def _get_cached_properties_names(self):
        return [name for name, value in inspect.getmembers(self.__class__)
                if isinstance(value, functools.cached_property) if name in self.__dict__.keys()]

    def delete_cached_properties(self):
        for cached_property in self._get_cached_properties_names():
            delattr(self, cached_property)
        self._dim = None

    def generators_eval(self, **kwargs):
        return [eval(generator.replace("^", "**"), kwargs) for generator in self.generators]

    @property
    def is_unit_ideal(self):
        return self == Ideal(self.ring, ('1', ))


def reduce(poly, ideal):
    return ideal.reduce(poly)


def monomial_to_exponents(variables, monomial):
    """Converts a monomial in the variables of a polynomial ring into a numpy.array of exponents."""
    """Converts a monomial in the variables of a polynomial ring into a numpy.array of exponents."""
    exps = numpy.zeros(len(variables), dtype=int)
    variables = list(map(str, variables))  # in case sympy symbols are passsed
    _split_monomial = list(filter(lambda x: x != '', monomial.replace("^", "**").split("*")))
    split_monomial = []
    for entry in _split_monomial:  # need to fix Singular notation for exponents (e.g. if x2 is not a ring variable then x^2 is written as x2)
        if not entry.isdigit() and entry not in variables:
            exp = re.findall(r"(\d+)", entry)[0]
            split_monomial += [entry.replace(exp, ""), exp]
        else:
            split_monomial += [entry]
    for i, ientry in enumerate(split_monomial):
        if not ientry.isdigit():
            if i < len(split_monomial) - 1 and split_monomial[i + 1].isdigit():
                exp = int(split_monomial[i + 1])
            else:
                exp = 1
            exps[variables.index(ientry)] += exp
    return exps



TIMEOUT = 60  # seconds  # noqa
DEGBOUND = 0  # 0 = no-bound  # noqa
DEGBOUNDs = [4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24]  # noqa
DEBUG = False  # noqa

from .version import __version__  # noqa
from .ideal import Ideal  # noqa
from .ring import Ring  # noqa
from .qring import QuotientRing, QRing  # noqa
from .tools import SingularException, Singular_version  # noqa
from .field import Field  # noqa
from .polynomial import Polynomial, Monomial  # noqa
from .settings import TemporarySetting  # noqa



import sympy
import re

from random import randint
from copy import deepcopy
from packaging.version import Version

from .tools import execute_singular_command, Singular_version
from .ring import Ring
from .settings import TemporarySetting


class Ideal_Algorithms:

    singular_commands_EXT = r"""
        // option(redSB);
        ring r = {r};
        ideal i = {ideal};
        // Block Order X \ U > U in F[X]
        ring r1 = {r1};
        ideal j = imap(r, i);
        ideal gr = groebner(j);
        // to F(U)[X \ U]
        ring r2 = {r2};
        ideal G = imap(r1, gr);
        print(G);
    """

    def extension(self, U, ordering="lp"):
        r"""Extension to K(U)[X \ U]"""
        r = self.ring
        X = r.variables
        XqU = tuple(entry for entry in X if entry not in U)
        if ordering == "lp":
            r1 = Ring('0', XqU + U, 'lp')
            r2 = Ring((sympy.symbols('0'), ) + U, XqU, 'lp')
        elif ordering == "dp":
            r1 = Ring('0', XqU + U, (f'dp({len(XqU)})', f'dp({len(U)})'))
            r2 = Ring((sympy.symbols('0'), ) + U, XqU, 'dp')
        else:
            raise ValueError
        string = execute_singular_command(self.singular_commands_EXT.format(**{"ideal": self, "r": r, "r1": r1, "r2": r2}))
        generators = string.split(",\n")
        Ideal = self.__class__
        I_extended = Ideal(r2, generators)
        return I_extended

    singular_commands_EXTCONT1 = r"""
        LIB "polylib.lib";
        // K(U)[X \ U]
        ring r2 = {r2};
        ideal G = {extended_ideal};
        // List of leadcoefficients
        list l;
        for (int i=1; i<=size(G); i=i+1)
        {{
            l = insert(l, leadcoef(G[i]));
        }}
        ring r = {r};
        list l = imap(r2, l);
        // f-polynomial
        list factors;
        for (i=1; i<=size(l); i=i+1)
        {{
            factors = insert(factors, factorize(l[i], 1));
        }}
        print("factors:"); print(factors);
        $
    """

    def extension_contraction_fpoly(self, U, ordering="lp"):
        """Returns a minimal list of factors for the f-polynomial in CONT and EXTCONT"""
        r = self.ring
        X = r.variables
        XqU = tuple(entry for entry in X if entry not in U)
        if ordering == "lp":
            r1 = Ring('0', XqU + U, 'lp')
            r2 = Ring((sympy.symbols('0'), ) + U, XqU, 'lp')
        elif ordering == "dp":
            r1 = Ring('0', XqU + U, (f'dp({len(XqU)})', f'dp({len(U)})'))
            r2 = Ring((sympy.symbols('0'), ) + U, XqU, 'dp')
        else:
            raise ValueError
        string = execute_singular_command(self.singular_commands_EXTCONT1.format(**{"extended_ideal": self.extension(U, ordering), "r": r, "r1": r1, "r2": r2}))
        f_polys_factors = list(set([row.split("=")[1].replace(" ", "") for row in string.split("\n") if "factors" not in row and "=" in row]))
        f_polys_factors = [entry for entry in f_polys_factors if entry != '1']
        return ['1'] + f_polys_factors

    singular_commands_EXTCONT2 = """
        LIB "polylib.lib";
        // option(redSB);
        ring r2 = {r2};
        ideal G = {extended_ideal};
        ring r = {r};
        ideal Iec = imap(r2, G);
        list factors = {f_polys_factors};
        list Iec_and_s;
        int s = 0;
        for (int i_=1; i_<=size(factors); i_=i_+1)
        {{
            Iec_and_s = {sat}(Iec, ideal(factors[i_]));
            Iec = Iec_and_s[1];
            s = max(s, Iec_and_s[2]);
        }}
        print("saturation index:"); print(s);
        print("extended-contracted ideal:"); print(minbase(Iec));
    """

    def extension_contraction(self, U, ordering="lp"):
        """Computes extension-contraction of self to localization define by indepSet U. Returns a tuple: (saturation index, extended-contracted ideal)."""
        r = self.ring
        X = r.variables
        XqU = tuple(entry for entry in X if entry not in U)
        if ordering == "lp":
            # r1 = Ring('0', XqU + U, 'lp')
            r2 = Ring((sympy.symbols('0'), ) + U, XqU, 'lp')
        elif ordering == "dp":
            # r1 = Ring('0', XqU + U, (f'dp({len(XqU)})', f'dp({len(U)})'))
            r2 = Ring((sympy.symbols('0'), ) + U, XqU, 'dp')
        else:
            raise ValueError
        string = execute_singular_command(self.singular_commands_EXTCONT2.format(**{
            "extended_ideal": self.extension(U, ordering), "f_polys_factors": ','.join(self.extension_contraction_fpoly(U, ordering)),
            "r": r, "r2": r2, "sat": "sat" if Singular_version <= Version('4.3.1') else 'sat_with_exp'}))
        Ideal = self.__class__
        return int(string.split("\n")[1]), Ideal(r, [entry.replace(",", "") for entry in string.split("\n")[3:]])

    def primeTestDLP(self, verbose=False, timeout_fpoly=10, timeout_dim=600,
                     iterated_degbound_computation=False, projection_number=None, astuple=False):
        """Returns True if the ideal is prime, False if it is not.
        If astuple is set to True, return (is_primary, is_prime) bool pair.
        Raises Inconclusive if it can't decide. Assumes equidimensionality of input ideal.
        Experimental new feature with iterated_degbound_computation=True,
        may help when ideal is primary and deg-unbounded computation fails.
        """
        import syngular
        # algorithm works over rings, if in a qring convert to the full ring.
        self = deepcopy(self)
        self.to_full_ring()
        # first step: find zero dimensional projections which are linear in the dependent variables, i.e. are a single point.
        if projection_number is not None:
            self.indepSets = self.indepSets[projection_number:projection_number + 1]
        lowest_degree_projection_indepSets = []
        lowest_degree = 9999999
        for i, indepSet in enumerate(self.indepSets):
            if verbose:
                print(f"\r@{i}/{len(self.indepSets)}", end="")
            X = self.ring.variables                                            # all variables: X
            U = tuple(var for is_indep, var in zip(indepSet, X) if is_indep)   # independent variables: U
            XqU = tuple(entry for entry in X if entry not in U)                # dependent variables X / U
            prime = 536870909  # largest prime below 2 ** 29
            projection = {entry: randint(1, prime) for entry in U}             # random values for U
            zeroDimSelf = deepcopy(self)
            zeroDimSelf.generators = [str(sympy.poly(entry.subs(projection), modulus=prime).as_expr())
                                      for entry in sympy.sympify(zeroDimSelf.generators)]
            zeroDimSelf.ring = Ring(str(prime), XqU, 'lp')
            zeroDimSelf.delete_cached_properties()
            degrees = list(map(int, re.findall(r"\^(\d)", "".join(zeroDimSelf.groebner_basis))))
            if degrees == []:
                max_degree = 1
            else:
                max_degree = max(degrees)
            if max_degree < lowest_degree:
                lowest_degree = max_degree
                lowest_degree_projection_indepSets = []
            if max_degree == lowest_degree:
                lowest_degree_projection_indepSets += [indepSet]
            if verbose:
                print(f"\r@{i}/{len(self.indepSets)} projections found: {len(lowest_degree_projection_indepSets)} of degree {lowest_degree}", end="")
        if verbose:
            print(f"\rprojections found: {len(lowest_degree_projection_indepSets)} of degree {lowest_degree}", end="                    ")
        # for each single-point zero-dimensional projection, get the factors of the f-polynomial in EXTCONT (Becker et al. Proposition 8.96)
        f_polys_factors = []
        for i, indepSet in enumerate(lowest_degree_projection_indepSets):
            if verbose:
                number_of_timedout_fpolys = ["TIMEDOUT" in entry[-1] for entry in f_polys_factors].count(True)
                print(f"\rgathering f-poly factors: @ {i}/{len(lowest_degree_projection_indepSets)} of which {number_of_timedout_fpolys} timedout",
                      end="                                   ")
            # 0 dim slice
            X = self.ring.variables
            U = tuple(var for is_indep, var in zip(indepSet, X) if is_indep)
            try:
                with TemporarySetting("syngular", "TIMEOUT", timeout_fpoly):
                    f_polys_factors += [self.extension_contraction_fpoly(U, "lp")]
            except TimeoutError:
                f_polys_factors += [['- TIMEDOUT - probably very very long ' * 8000]]
                continue
            except Exception as e:
                if verbose:
                    print(f"An exception occurred: {e}")
                f_polys_factors += [['- TIMEDOUT - probably very very long ' * 8000]]
                continue
        # just keep the one with the smallest greatest factor
        max_lengths = [max(map(len, f_poly_factors)) for f_poly_factors in f_polys_factors]
        easiest_projection_indepSet = lowest_degree_projection_indepSets[max_lengths.index(min(max_lengths))]
        easiest_projection_indepSet_index = self.indepSets.index(easiest_projection_indepSet)
        if verbose:
            print(f"\neasiest projection is {easiest_projection_indepSet_index}: {easiest_projection_indepSet} of degree {lowest_degree}")
        smallest_fpoly_factors = f_polys_factors[max_lengths.index(min(max_lengths))]
        if lowest_degree > 1:
            tries_for_irreducibility = 100
            for i in range(tries_for_irreducibility):
                # Warning repeated code from above, use _semi_numerical_slice?
                indepSet = easiest_projection_indepSet
                X = self.ring.variables                                            # all variables: X
                U = tuple(var for is_indep, var in zip(indepSet, X) if is_indep)   # independent variables: U
                XqU = tuple(entry for entry in X if entry not in U)                # dependent variables X / U
                prime = 536870909  # largest prime below 2 ** 29
                projection = {entry: randint(1, prime) for entry in U}             # random values for U
                zeroDimSelf = deepcopy(self)
                zeroDimSelf.generators = [str(sympy.poly(entry.subs(projection), modulus=prime).as_expr())
                                          for entry in sympy.sympify(zeroDimSelf.generators)]
                zeroDimSelf.ring = Ring(str(prime), XqU, 'lp')
                zeroDimSelf.delete_cached_properties()
                zeroDim_prim_dec = zeroDimSelf.primary_decomposition
                if len(zeroDim_prim_dec) == 1:
                    Q, P = zeroDim_prim_dec[0]
                    radical = True if Q == P else False
                    print("\rWas irreducible at the chosen point. Continuing test.", end="                    ")
                    break  # conclusive
                elif verbose:
                    print(f"\rWas reducible at the chosen points, checking again. At try {i}.", end="")
            else:
                raise Inconclusive(f"Likely reducible, since it was reducible on {tries_for_irreducibility} zero dim extensions.")
                # return False if not astuple else (False, False)   - statistical statement, can't be 100% sure
        else:
            radical = True
        if smallest_fpoly_factors == ['- TIMEDOUT - probably very very long ' * 8000]:
            raise Inconclusive("Timedout on fpoly factors gathering.")
        smallest_fpoly_factors = sorted(smallest_fpoly_factors, key=lambda x: len(x))
        if verbose:
            print(f"\r smallest f poly factors ({len(smallest_fpoly_factors)}): complexity {sum(map(len, smallest_fpoly_factors))}",
                  end="                              \n")
        # check that the dimensionality drops when adding each of these factors separately (and hence drops for <ideal, f^s>)
        with TemporarySetting("syngular", "TIMEOUT", timeout_dim):
            self.codim  # just cache it
            for i, factor in enumerate(smallest_fpoly_factors):
                if verbose:
                    print(f"\r at factor {i}: {factor}.", end="                                       \n")
                X = deepcopy(self)
                X.generators += [factor]
                X.delete_cached_properties()
                # Experimental - Assumes codim w/ deg bound <= true codim.
                # Helps termiante the prime test early, IF the result is True.
                for deg_bound in syngular.DEGBOUNDs * iterated_degbound_computation:
                    with TemporarySetting("syngular", "DEGBOUND", deg_bound):
                        X.delete_cached_properties()
                        if self.codim < X.codim:
                            if verbose:
                                print(f"deg_bound {deg_bound} computation was conclusive.", end="\n")
                            break
                        else:
                            if verbose:
                                print(f"deg_bound {deg_bound} computation was inconclusive.", end="\n")

                else:  # loop completed without encountering a break
                    if verbose:
                        print("deg_bound reset to zero. Performing full computation.", end="\n")
                    with TemporarySetting("syngular", "DEGBOUND", 0):
                        X.delete_cached_properties()
                        # if self.indepSet.count(0) >= X.indepSet.count(0):   # deprecated, equivalent to next line.
                        if self.codim >= X.codim:
                            return False if not astuple else (False, False)
        return radical if not astuple else (True, radical)

    def test_primality(self, *args, **kwargs):
        """Tests if an ideal is prime. Not to be confused with primary."""
        # eventually check assumptions and decide which test to use
        return self.primeTestDLP(*args, **kwargs)


class Inconclusive(Exception):
    pass



import syngular

from .ring import Ring


class QuotientRing(Ring):

    def __init__(self, ring, ideal):
        super().__init__(ring.field, ring.variables, ring.ordering)
        self.ideal = ideal

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        string = f"{super().__str__()};\nideal i_ = {self.ideal};\nqring q = std(i_)"
        if syngular.DEGBOUND != 0:
            string += f";\ndegBound = {syngular.DEGBOUND};\noption()"
        return string

    def __repr__(self):
        return str(self)


QRing = QuotientRing



import importlib


class TemporarySetting:
    def __init__(self, module_name, setting_name, new_value):
        self.module_name = module_name
        self.setting_name = setting_name
        self.new_value = new_value
        self.old_value = None

    def __enter__(self):
        module = importlib.import_module(self.module_name)
        if hasattr(module, self.setting_name):
            self.old_value = getattr(module, self.setting_name)
            setattr(module, self.setting_name, self.new_value)
        else:
            raise AttributeError(f"Setting '{self.setting_name}' does not exist in module '{self.module_name}'")

    def __exit__(self, exc_type, exc_val, exc_tb):
        module = importlib.import_module(self.module_name)
        if hasattr(module, self.setting_name):
            setattr(module, self.setting_name, self.old_value)



from setuptools import setup, find_packages
from pathlib import Path


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


with (this_directory / "syngular" / "version.py").open() as f:
    exec(f.read())
    version = __version__  # noqa


setup(
    name='syngular',
    version=version,
    license='GNU General Public License v3.0',
    description='An Object-Oriented Python Interface and Extension to Singular',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Giuseppe De Laurentis',
    author_email='g.dl@hotmail.it',
    url='https://github.com/GDeLaurentis/syngular',
    download_url=f'https://github.com/GDeLaurentis/syngular/archive/{version}.tar.gz',
    project_urls={
        'Documentation': 'https://gdelaurentis.github.io/syngular/',
        'Issues': 'https://github.com/GDeLaurentis/syngular/issues',
    },
    keywords=['syngular', 'singular', 'algebraic geometry', 'ring', 'ideal',
              'polynomials', 'polynomial equations'],
    packages=find_packages(),
    install_requires=[
        'numpy',
        'sympy',
        'pyadic',
        'multiset'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)



import pytest
import sympy
import numpy

from syngular import Ideal, Ring, SingularException
from syngular.ideal import monomial_to_exponents, reduce


def test_ideal_instantiation():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1*x2'])
    assert I is not None


def test_invalid_ideal_instantiation():
    with pytest.raises(SingularException):
        Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x*y'])


def test_ideal_quotient():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1*x2'])
    J = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2*x2'])
    K = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1'])
    R = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['1'])
    assert R / I == R
    assert I / R == I
    assert J / I == K


def test_ideal_mul_and_power():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1', 'x2'])
    J = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2', 'x1*x2', 'x2^2'])
    R = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['1'])
    assert I ** 2 == I * I == J
    assert I ** 0 == J ** 0 == R


def test_primary_decomposition():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2*x2'])
    J = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x2'])
    K = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2'])
    L = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1'])
    assert I.primary_decomposition == [(J, J), (K, L)]


def test_radical():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2*x2'])
    J = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x2'])
    L = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1'])
    assert I.radical == J & L


def test_primary_intersection():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2*x2'])
    J = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x2'])
    K = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2'])
    assert I == J & K
    assert I == Ideal.intersection(*(J, I, K))
    assert I == Ideal.intersection(*(I, ))


def test_eliminate():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1', 'x2'])
    J = Ideal(Ring('0', ('x1', ), 'dp'), ['x1'])
    K = Ideal(Ring('0', ('x2', ), 'dp'), ['x2'])
    assert I.eliminate(range(1, 1)) == K
    assert I.eliminate(range(2, 2)) == J


def test_add():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1', 'x2'])
    J = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1'])
    K = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x2'])
    assert I == J + K


def test_poly_contains():
    K = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2'])
    L = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1'])
    assert 'x1' not in K
    assert 'x1' in L
    assert sympy.symbols('x1') in L


def test_ideal_contains():
    K = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2'])
    L = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1'])
    assert K in L
    assert L not in K


def test_indepSet():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2*x2'])
    assert I.indepSet in [(1, 0), (0, 1)]


def test_indepSets():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2*x2'])
    assert I.indepSets == [(1, 0), (0, 1)]


def test_hash_and_set():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2*x2'])
    J = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x2'])
    assert {I, I} == {I}
    assert {I, I, J} == {I, J, J} == {I, J}


def test_ideal_to_qring_and_back():
    ring = Ring('0', ('x1', 'x2'), 'dp')
    I = Ideal(ring, ['x1', 'x2'])
    J = Ideal(ring, ['x1'])
    I.to_qring(J)
    assert I.groebner_basis == I.minbase == ['x2']
    I.to_full_ring()
    assert set(I.groebner_basis) == set(I.minbase) == set(['x1', 'x2'])


def test_lead_gb_monomials():
    R = Ring('0', ('x', 'y', ), 'dp')
    I = Ideal(R, ['x^2', 'y'])
    assert I.leadGBmonomials == ['y', 'x^2']


def test_ideal_reduce_poly():
    ring = Ring('0', ('x1', 'x2'), 'dp')
    I = Ideal(ring, ['x1'])
    poly = 'x1+x2'
    assert I.reduce(poly) == 'x2' == reduce(poly, I)


def test_ideal_reduce_ideal():
    ring = Ring('0', ('x1', 'x2'), 'dp')
    I = Ideal(ring, ['x1', 'x2'])
    J = Ideal(ring, ['x1'])
    K = J.reduce(I)
    assert K == Ideal(ring, ['x2'])
    L = I.reduce(J)
    assert L == Ideal(ring, ['0'])


def test_ideal_dim():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), [])
    assert I.dim == 2
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1*x2'])
    assert I.dim == 1


def test_ideal_dims_and_codims():
    I = Ideal(Ring('0', ('x1', 'x2', 'x3'), 'dp'), ['(x3+1)*x1', '(x2+1)*x1'])
    assert I.dim == 2
    assert I.dims == {1, 2}
    assert I.codim == 1
    assert I.codims == {1, 2}


def test_monomial_to_exponents():
    r = Ring('0', ('x1', 'x2', 'x3'), 'dp')
    monomial = 'x1*x2^4*x3^123'
    assert numpy.all(monomial_to_exponents(r.variables, monomial) == numpy.array([1, 4, 123]))


def test_monomial_to_exponents_Singular_notation():
    r = Ring('0', ('x', 'y', 'z'), 'dp')
    monomial = 'x*y4*z123'
    assert numpy.all(monomial_to_exponents(r.variables, monomial) == numpy.array([1, 4, 123]))



import pytest
import syngular

from syngular import Ideal, Ring
# from syngular.ideal_algorithms import Inconclusive


def test_extension_contraction():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1*x2'])
    J = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x2'])
    indepSet = I.indepSets[0]
    X = I.ring.variables
    U = tuple(var for is_indep, var in zip(indepSet, X) if is_indep)
    assert I.extension_contraction(U, ordering="lp") == (1, J)
    assert I.extension_contraction(U, ordering="dp") == (1, J)
    with pytest.raises(ValueError):
        I.extension_contraction(U, ordering="Dp") == (1, J)


def test_test_primality():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1*x2'])
    J = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x2'])
    assert I.test_primality(verbose=True) is False
    assert I.test_primality(verbose=True, iterated_degbound_computation=True) is False
    assert J.test_primality(verbose=True) is True
    assert J.test_primality(verbose=True, iterated_degbound_computation=True) is True


def test_test_primality_primary_but_not_prime():
    I = Ideal(Ring('0', ('x1', 'x2'), 'dp'), ['x1^2'])
    assert I.test_primality(verbose=True) is False
    assert I.test_primality(verbose=True, iterated_degbound_computation=True) is False
    assert I.test_primality(verbose=True, astuple=True) == (True, False)
    assert I.test_primality(verbose=True, iterated_degbound_computation=True, astuple=True) == (True, False)


def test_test_primality_with_degree_bound():
    # from: LipsIdeal(6, ('⟨1|2+3|1]', '⟨2|1+4|2]'))
    r = Ring(0, ('d6', 'c6', 'b6', 'a6', 'd5', 'c5', 'b5', 'a5', 'd4', 'c4', 'b4', 'a4', 'd3', 'c3', 'b3', 'a3', 'd2', 'c2', 'b2', 'a2', 'd1', 'c1', 'b1', 'a1'), 'dp')
    generators = ['4*a1*b2*c1*d2 - 4*a1*b2*c2*d1 + 4*a1*b3*c1*d3 - 4*a1*b3*c3*d1 - 4*a2*b1*c1*d2 + 4*a2*b1*c2*d1 - 4*a3*b1*c1*d3 + 4*a3*b1*c3*d1',
                  '4*a1*b2*c1*d2 - 4*a1*b2*c2*d1 - 4*a2*b1*c1*d2 + 4*a2*b1*c2*d1 + 4*a2*b4*c2*d4 - 4*a2*b4*c4*d2 - 4*a4*b2*c2*d4 + 4*a4*b2*c4*d2',
                  'b1*d1 + b2*d2 + b3*d3 + b4*d4 + b5*d5 + b6*d6',
                  '-a1*d1 - a2*d2 - a3*d3 - a4*d4 - a5*d5 - a6*d6',
                  '-b1*c1 - b2*c2 - b3*c3 - b4*c4 - b5*c5 - b6*c6',
                  'a1*c1 + a2*c2 + a3*c3 + a4*c4 + a5*c5 + a6*c6']
    I = Ideal(r, generators)
    syngular.DEGBOUNDs = [4, 6, ]
    assert I.test_primality(verbose=True, iterated_degbound_computation=True, projection_number=192)



#!/usr/bin/env python
import re

repo_owner = "GDeLaurentis"
repo_name = "syngular"


def update_coverage_badge():
    """Updates the coverage badge in README.md"""
    # read value from coverage.txt
    with open('coverage.txt', 'r') as f:
        content = f.read()
    content = content.split("TOTAL   ")[1].split("\n\n")[0]
    coverage_percentage = int(re.findall(r"(\d+)%", content)[0])
    # obtain colour of badge
    coverage_steps = {
        100: "brightgreen",
        90: "green",
        80: "greenyellow",
        70: "yellow",
        50: "orange",
        0: "red",
    }
    for key in sorted(coverage_steps.keys())[::-1]:
        if coverage_percentage >= key:
            break
    coverage_colour = coverage_steps[key]
    # update README.md
    with open('README.md', 'r') as f:
        lines = f.readlines()
    for i, line in enumerate(lines):
        if "![Coverage]" in line:
            lines[i] = '[![Coverage](https://img.shields.io/badge/Coverage-{}%25-{}?labelColor=2a2f35)](https://github.com/{}/{}/actions)\n'.format(
                coverage_percentage, coverage_colour, repo_owner, repo_name)
    with open('README.md', 'w') as f:
        f.write("".join(lines))


if __name__ == "__main__":
    update_coverage_badge()



import functools
import numpy
import operator
import re
import sympy

from collections import defaultdict
from multiset import FrozenMultiset
from fractions import Fraction as Q
from copy import deepcopy

from .field import Field


class Monomial(FrozenMultiset):
    """A FrozenMultiset representation of a Monomial. Positive integer multiplicities represent powers."""

    def __init__(self, letters_and_powers={}):
        if isinstance(letters_and_powers, (dict, tuple, FrozenMultiset, Monomial)):
            super(Monomial, self).__init__(letters_and_powers)
        elif isinstance(letters_and_powers, str):
            super(Monomial, self).__init__(self.__rstr__(letters_and_powers))
        else:
            print("entry:", repr(letters_and_powers))
            print("type:", type(letters_and_powers))
            raise NotImplementedError

    @staticmethod
    def __rstr__(monomial):
        if monomial == '':
            return dict()
        monomial = monomial.replace("**", "^")
        factors = monomial.split('*')
        factor_groups = defaultdict(list)
        for factor in factors:
            if '^' in factor:
                function, power = factor.split('^')
                power = int(power)
            else:
                function = factor
                power = 1
            factor_groups[function].append(power)
        powers = [(function, sum(powers)) for function, powers in factor_groups.items()]
        return dict(powers)

    def __repr__(self):
        return f"Monomial(\"{str(self)}\")"

    def __str__(self):
        return "*".join([f"{key}^{val}" for key, val in self.items()])

    def tolist(self):
        return [entry for key, val in self.items() for entry in [key, ] * val]

    def subs(self, values_dict):
        return functools.reduce(operator.mul, [values_dict[key] ** val for key, val in self.items()], 1)

    def __mul__(self, other):
        assert isinstance(other, Monomial)
        return super(Monomial, self).__add__(other)

    def __truediv__(self, other):
        raise Exception("Monomial division not implement. Do you mean this to be a Rational Function?")

    def __add__(self, other):
        raise Exception("Monomial addition not implement. Do you mean this to be a Polynomial?")

    def __sub__(self, other):
        raise Exception("Monomial subtraction not implement. Do you mean this to be a Polynomial?")


class Polynomial(object):
    """Generalization of the concept of Multiset where multiplicities are in an arbitrary Field and the elements are Monomials."""

    # This allows the intantiation step to return an element of the udnerlying field.
    # def __new__(cls, coeffs_and_monomials, field):
    #     self = super(Polynomial, cls).__new__(cls)
    #     self.__init__(coeffs_and_monomials, field)
    #     if len(self) == 1 and self.coeffs_and_monomials[0][1] == Monomial(""):
    #         return self.coeffs_and_monomials[0][0]
    #     return self

    def __init__(self, coeffs_and_monomials, field):
        if isinstance(coeffs_and_monomials, (str, int)) or isinstance(coeffs_and_monomials, sympy.Basic):
            coeffs_and_monomials = self.__rstr__(str(coeffs_and_monomials), field)
        elif isinstance(coeffs_and_monomials, Polynomial):
            coeffs_and_monomials = coeffs_and_monomials.coeffs_and_monomials
        elif isinstance(coeffs_and_monomials, list):
            assert all([isinstance(entry, tuple) for entry in coeffs_and_monomials])
            assert all([len(entry) == 2 for entry in coeffs_and_monomials])
            assert all([entry[0] in field for entry in coeffs_and_monomials if entry[0] != 0])
            assert all([isinstance(entry[1], Monomial) for entry in coeffs_and_monomials])
        elif coeffs_and_monomials in field:
            coeffs_and_monomials = [(coeffs_and_monomials, Monomial(''))]
        else:
            raise NotImplementedError(f"Received {coeffs_and_monomials} \n of type {type(coeffs_and_monomials)}")
        self.coeffs_and_monomials = coeffs_and_monomials
        self._field = field

    def __str__(self, for_repr=False):
        return " + ".join((f"({str(coeff) if not for_repr else repr(coeff)})"
                           if hasattr(self.field, "name") and self.field.name in ["padic", "finite field", ] else
                           f"{str(coeff) if not for_repr else repr(coeff)}") +
                          (f"*{monomial}" if str(monomial) != "" else "")
                          for coeff, monomial in self.coeffs_and_monomials).replace("+-", "-").replace("+ -", "- ")

    def __repr__(self):
        return f"Polynomial(\"{self.__str__(for_repr=True)}\", {self.field})"

    def __getitem__(self, index):
        return self.coeffs_and_monomials[index]

    @property
    def coeffs_and_monomials(self):
        return self._coeffs_and_monomials

    @coeffs_and_monomials.setter
    def coeffs_and_monomials(self, temp_coeffs_and_monomials):
        self._coeffs_and_monomials = [(coeff, monomial) for coeff, monomial in temp_coeffs_and_monomials if coeff != 0]
        self.reduce()
        if self._coeffs_and_monomials == []:  # ensure there is at least an entry
            self._coeffs_and_monomials = [(0, Monomial(''))]

    @property
    def field(self):
        return self._field

    @field.setter
    def field(self, temp_field):
        if not isinstance(temp_field, Field) and temp_field not in [int, Q]:
            raise Exception(f"temp_field: {temp_field} is not a field")
        if not hasattr(self, "field") or temp_field != self.field:
            self._field = temp_field
            self.coeffs_and_monomials = [(temp_field(coeff), monomial) for coeff, monomial in self.coeffs_and_monomials]

    @property
    def coeffs(self):
        return [coeff for coeff, monomial in self.coeffs_and_monomials]

    @coeffs.setter
    def coeffs(self, temp_coeffs):
        self.coeffs_and_monomials = [(temp_coeff, monomial) for temp_coeff, (coeff, monomial) in zip(temp_coeffs, self.coeffs_and_monomials)]

    @property
    def monomials(self):
        return [monomial for _, monomial in self.coeffs_and_monomials]

    @property
    def variables(self):
        return set([var for monomial in self.monomials for var in monomial.tolist()])

    def rationalise(self):
        from pyadic.finite_field import vec_chained_FF_rationalize
        rat_coeffs = vec_chained_FF_rationalize([numpy.array(self.coeffs).astype(int), ], [self.field.characteristic, ]).tolist()
        rat_poly = deepcopy(self)
        rat_poly.coeffs = rat_coeffs
        rat_poly.field = Q
        return rat_poly

    def reduce(self):
        """Merges equal monomials"""
        unequal_monoms = set([monom for _, monom in self.coeffs_and_monomials])
        if len(unequal_monoms) == len(self.coeffs_and_monomials):
            return  # already reduced
        new_coeffs_and_monomials = []
        for unequal_monom in unequal_monoms:
            merged_coeff = sum([coeff for coeff, monom in self.coeffs_and_monomials if monom == unequal_monom])
            new_coeffs_and_monomials += [(merged_coeff, unequal_monom)]
        self.coeffs_and_monomials = new_coeffs_and_monomials

    @staticmethod
    def __rstr__(polynomial, field):
        polynomial = polynomial.replace(" ", "").replace("+-", "-")
        polynomial = re.sub(r"(\+|\-)I\*{0,1}([\d\.e\+\-]+)", r"\1\2j", polynomial)  # format complex nbrs
        parentheses = [(("(", ), (")", )), (("{", ), ("}", )), (("[", "⟨", "<", ), ("]", "⟩", ">"))]
        lopen_parentheses = [parenthesis[0] for parenthesis in parentheses]
        lclos_parentheses = [parenthesis[1] for parenthesis in parentheses]
        parentheses_balance = [0 for _ in parentheses]
        next_match = ""
        coeffs_and_monomials_strings = []
        for char in polynomial:
            if (char == "+" or char == "-") and all([parenthesis_balance == 0 for parenthesis_balance in parentheses_balance]):
                if next_match != "":
                    coeffs_and_monomials_strings += [next_match.replace(" ", "").replace("**", "^")]
                next_match = char
            else:
                if any([char in open_parentheses for open_parentheses in lopen_parentheses]):
                    parentheses_balance[[char in open_parentheses for open_parentheses in lopen_parentheses].index(True)] += 1
                elif any([char in clos_parentheses for clos_parentheses in lclos_parentheses]):
                    parentheses_balance[[char in clos_parentheses for clos_parentheses in lclos_parentheses].index(True)] -= 1
                next_match += char
        else:
            assert all([parenthesis_balance == 0 for parenthesis_balance in parentheses_balance])
            coeffs_and_monomials_strings += [next_match.replace(" ", "").replace("**", "^")]
        # print(coeffs_and_monomials_strings)
        coeffs_and_monomials = []
        for coeff_and_monomial_string in coeffs_and_monomials_strings:
            coeff = re.findall(r"^[+-]?[\%\+\-\(\)\.j\d/e]*", coeff_and_monomial_string)[0]
            monomial = coeff_and_monomial_string.replace(coeff, "", 1)
            coeff = "+1" if coeff == "+" or coeff == "" else "-1" if coeff == "-" else coeff
            coeff_denom = re.findall(r"/(\d+)", monomial)
            if coeff_denom != []:
                coeff_denom = int(coeff_denom[0])
                monomial = monomial.replace(f"/{coeff_denom}", "")
            else:
                coeff_denom = 1
            if len(monomial) > 0 and monomial[0] == "*":
                monomial = monomial[1:]
            if coeff[:2] == "+(" and coeff[-1:] == ")":
                coeff = coeff[2:-1]
            # print("string:", coeff_and_monomial_string, "\ncoeff", coeff, "\nmonomial:", monomial)
            coeffs_and_monomials += [(field(coeff) / coeff_denom, Monomial(monomial))]
        return coeffs_and_monomials

    def subs(self, base_point, field=None):
        if field is None:
            field = self.field
        base_point = {str(key): val for key, val in base_point.items()}
        new_coeffs_and_monomials = []
        for coeff, monomial in self.coeffs_and_monomials:
            poly = Polynomial(coeff * monomial.subs(base_point), field)
            new_coeffs_and_monomials += [*(poly.coeffs_and_monomials)]
        new_poly = Polynomial(new_coeffs_and_monomials, field)
        new_poly.reduce()
        return new_poly

    def __call__(self, *args, **kwargs):
        return self.subs(*args, **kwargs)

    def __len__(self):
        return len(self.coeffs_and_monomials)

    def __eq__(self, other):
        if other == 0:
            if len(self) == 1 and self.coeffs[0] == 0 and self.monomials[0] == Monomial(''):
                return True
            else:
                return False
        elif isinstance(other, Polynomial):
            return set(self.coeffs_and_monomials) == set(other.coeffs_and_monomials)
        elif isinstance(other, str):
            try:
                return set(self.coeffs_and_monomials) == set(Polynomial(other, self.field).coeffs_and_monomials)
            except:  # noqa
                return False
        else:
            raise NotImplementedError(f"Operation: __eq__; self: {self}; self class {self.__class__}; other: {other}; other class {other.__class__}.")

    def __truediv__(self, other):
        return Polynomial([(coeff / other, monom) for coeff, monom in self.coeffs_and_monomials], self.field)

    def __add__(self, other):
        if isinstance(other, Polynomial):
            return Polynomial(self.coeffs_and_monomials + other.coeffs_and_monomials, self.field)
        else:
            raise NotImplementedError(f"Operation: __add__; self: {self}; self class {self.__class__}; other: {other}; other class {other.__class__}.")

    def __sub__(self, other):
        return self + (-1 * other)

    def __mul__(self, other):
        if isinstance(other, Polynomial):
            new_coeffs_and_monomials = []
            for coeff1, monomial1 in self.coeffs_and_monomials:
                for coeff2, monomial2 in other.coeffs_and_monomials:
                    new_coeffs_and_monomials += [(coeff1 * coeff2, monomial1 * monomial2)]
            return Polynomial(new_coeffs_and_monomials, self.field)
        elif isinstance(other, (int, Q)) or other in self.field:
            new_coeffs_and_monomials = [(other * coeff, monomial) for coeff, monomial in self.coeffs_and_monomials]
            return Polynomial(new_coeffs_and_monomials, self.field)
        else:
            raise NotImplementedError(f"Operation: __mul__; self: {self}; self class {self.__class__}; other: {other}; other class {other.__class__}.")

    def __rmul__(self, other):
        return self * other

    def __neg__(self):
        return -1 * self

    def __pow__(self, n):
        assert (isinstance(n, int) or n.is_integer())
        if n < 0:
            raise Exception("Polynomial to negative power is Rational Function.")
        elif n == 0:
            return Polynomial('1', self.field)
        elif n % 2 == 0:
            root_2_res = self ** (n / 2)
            return root_2_res * root_2_res
        else:
            return self * (self ** (n - 1))
